# -*- coding: utf-8 -*-

"""Tools for working with collection types."""


import logging
from collections import defaultdict

from typing import Any
from typing import Union
from typing import Callable
from typing import Optional

from collections.abc import Iterable
from collections.abc import Generator
from collections.abc import Collection


log = logging.getLogger(__name__)


#
#   COLLECTIONS
#

A, B, C, D = Any, Any, Any, Any
Index = int


# python 3.10
# def buckets(
#     col: Collection[A] | Collection[tuple[B, C]],
#     key: Optional[Callable[[Index, A], tuple[B, C]]] = None,
#     mapper: Optional[Callable[[tuple[C]], D]] = None,
# ) -> dict[B, list[C]] | dict[B, D]:
def buckets(
    col: Union[Collection[A], Collection[tuple[B, C]]],
    key: Optional[Callable[[Index, A], tuple[B, C]]] = None,
    mapper: Optional[Callable[[tuple[C]], D]] = None,
) -> Union[dict[B, list[C]], dict[B, D]]:
    """
    Sort data into buckets.

    Takes a collection and sorts the data into buckets based on a
    provided function. The resulting buckets can then optionally be
    mapped (e.g. to be reduced).

    Parameters
    ----------
    col : Union[Collection[A], Collection[tuple[B, C]]]
        Collection to be partitioned
    key : Optional[Callable[[Index, A], tuple[B, C]]]
        Optional function that returns (key, value) tuples
    mapper : Optional[Callable[[tuple[C]], D]]
        Optional function that takes a bucket and maps it

    Returns
    -------
    Union[dict[B, list[C]], dict[B, D]]
        A dictionary which maps bucket identifieres to their data

    Examples
    --------
    >>> from ktz.collections import buckets
    >>> data = [1, 2, 3, 4, 5]
    >>> buckets(col=data, key=lambda i, x: (x % 2 ==0 , x))
    {False: [1, 3, 5], True: [2, 4]}
    >>> buckets(col=data, key=lambda i, x: (x % 2 ==0 , x), mapper=sum)
    {False: 9, True: 6}

    """
    dic = defaultdict(list)

    for i, elem in enumerate(col):
        k, v = elem if key is None else key(i, elem)
        dic[k].append(v)

    if mapper:
        dic = {k: mapper(v) for k, v in dic.items()}

    return dict(dic)


def unbucket(
    buckets: dict[A, list[B]],
) -> list[tuple[A, B]]:
    """
    Flattens a bucket dictionary.

    Partitioned data is joined back up together and
    presented as tuples in a flat list.

    Parameters
    ----------
    buckets : dict[A, list[B]]
        Bucket dictionary

    Returns
    -------
    list[tuple[A, B]]
        Flattened collection

    Examples
    --------
    >>> from ktz.collections import buckets
    >>> from ktz.collections import unbucket
    >>> data = [1, 2, 3, 4, 5]
    >>> kcol.buckets(col=data, key=lambda i, x: (x % 2 ==0 , x))
    {False: [1, 3, 5], True: [2, 4]}
    >>> parts = kcol.buckets(col=data, key=lambda i, x: (x % 2 ==0 , x))
    >>> unbucket(parts)
    [(False, 1), (False, 3), (False, 5), (True, 2), (True, 4)]

    """
    return [(key, el) for key, lis in buckets.items() for el in lis]


Nested = Union[Iterable[A], A]


def flat(
    col: Iterable[Nested],
    depth: int = -1,
) -> Generator[A, None, None]:
    """Flattens a collection.

    Consumes the given iterable and flattens it up to n level deep or
    completely.

    Parameters
    ----------
    col : Iterable[Nested]
        Nested collection
    depth : int
        Maximum depth to flatten

    Returns
    -------
    Generator[A, None, None]
        Generator with flattened collection

    Examples
    --------
    >>> from ktz.collections import flat
    >>> flat([[1], [[2]]], depth=2)
    <generator object flat at 0x7f2886aeccf0>
    >>> list(flat([[1], [[2]]], depth=2))
    [1, [2]]

    """
    if depth == 0:
        yield col
        return

    try:
        for elem in col:
            yield from flat(col=elem, depth=depth - 1)
    except TypeError:
        yield col



# -*- coding: utf-8 -*-

"""String operations."""


import hashlib
from itertools import zip_longest

from typing import Any
from typing import Union
from typing import Optional
from typing import Callable

from collections.abc import Iterable


A = Any


def _apply_fns(col, fn, fns):
    col = list(col)

    if fn is not None:
        fns = [fn] * len(col)

    if fns is not None and len(fns):
        trans = []
        for fn, x in zip_longest(fns, col):
            trans.append(x if fn is None else fn(x))
        col = trans

    return col


def decode_line(
    encoded: bytes,
    sep: str = ",",
    fn: Callable[[A], str] = None,
    fns: Iterable[Optional[Callable[[str], A]]] = None,
) -> tuple[Union[str, A]]:
    """
    Decode a value list bytestring.

    Takes a unicode encoded bytestring separated by the seperator
    token(s) and returns a value tuple. If fns are given, these are
    used to transform the separated values.

    Parameters
    ----------
    encoded : bytes
        The value bytestring
    sep : str
        Separator token
    fn  : Optional[Callable[[A], str]]
        Optional converter applied to all, overrides fns
    fns : Iterable[Optional[Callable[[str], A]]]
        Optional converter functions

    Returns
    -------
    tuple[str | A]
        The separated values

    Examples
    --------
    from ktz.string import decode_line
    >>> line = "Hellö | 22 | True".encode("unicode_escape")
    >>> decode_line(line, sep="|", fns=(str, int, bool))
    ('Hellö', 22, True)

    """
    mapped = list(map(str.strip, encoded.decode("unicode_escape").split(sep)))
    mapped = _apply_fns(col=mapped, fn=fn, fns=fns)

    return tuple(mapped)


def encode_line(
    data: Iterable[Union[str, A]],
    sep: str = ",",
    fn: Callable[[A], str] = None,
    fns: Iterable[Optional[Callable[[A], str]]] = None,
) -> bytes:
    r"""
    Encode a value collection.

    Take a collection of either string values or other (requires fns)
    and produces a bytestring representation with values separated by
    sep. If fns are set, values are mapped accordingly. Adds a newline.

    Parameters
    ----------
    data : Iterable[str | A]
        Collection to be encoded
    sep : str
        Separator token
    fn  : Optional[Callable[[A], str]]
        Optional converter applied to all, overrides fns
    fns : Iterable[Optional[Callable[[A], str]]]
        Optional converter functions

    Returns
    -------
    bytes
        Encoded bytestring

    Examples
    --------
    >>> from ktz.string import decode_line
    >>> line = encode_line(("Hellö", 22, True), sep="|", fn=str)
    >>> line
    b'Hell\\xf6|22|True\n'
    >>> decode_line(line, sep="|", fns=(str, int, bool))
    ('Hellö', 22, True)

    """
    data = _apply_fns(col=data, fn=fn, fns=fns)
    assert all(sep not in s for s in data)

    return ((sep).join(data)).encode("unicode_escape") + b"\n"


def args_hash(
    *args,
) -> str:
    """
    Produce a hash value for the provided args.

    All provided arguments are transformed by their __str__
    implementation.

    Parameters
    ----------
    *args : To be hashed arguments

    Returns
    -------
    str
        Hash value for the arguments

    Examples
    --------
    >>> from ktz.string import args_hash
    >>> args_hash(10, "foo", True)
    'a8c83023141d8ed54fcb3d019bdf61a7468f01ac5704d0eb5bf1c626'
    """
    bytestr = "".join(map(str, args)).encode()
    return hashlib.sha224(bytestr).hexdigest()



# -*- coding: utf-8 -*-

"""
Extensions for dataclasses.

This module offers
  Index: an inverted index for dataclasses
  Builder: to iteratively build (immutable) data objects
"""

import ktz

from dataclasses import fields
from dataclasses import asdict
from collections import defaultdict

from typing import Any
from typing import Generic
from typing import TypeVar
from collections.abc import Iterable

# must be a dataclass
# I omit the somwhat wonky type checking for that
T = TypeVar("T")


class Index(Generic[T]):
    """
    Maintain an inverted index for dataclasses.

    Provided instances are saved to indexes based on their
    properties. The can then be retrieved fast by those properties.

    Methods
    -------
    get:
      Retrieve an object by key
    dis:
      Retrieve the union result for multi-key queries
    con:
      Retrieve the intersection result for multi-key queries

    """

    @property
    def flat(self) -> set[T]:
        """Return all indexed dataclasses.

        Note that this does not return a copy. You must create a copy
        before modifying this datastructure, otherwise you can expect
        undefined behaviour.

        Returns
        -------
        set[T]
            Set of all indexed dataclasses

        """
        return self._flat

    # ---

    def _raise_keyerror(self, key):
        raise KeyError(f"Field '{key}' not indexed.")

    def _agg(self, **kwargs) -> list[set[T]]:
        agg = []
        for key, val in kwargs.items():
            if key not in self._idxs:
                self._raise_keyerror(key)

            sub = set()
            for val in val if type(val) is set else {val}:
                sub |= self._idxs[key].get(val, set())

            agg.append(sub)

        return agg

    def get(self, **kwargs) -> set[T]:
        """
        Shortcut for single-field queries.

        If a single field is given, no distinction between dis() and
        con() exists. So this is a nice shortcut for simple queries.

        Parameters
        ----------
        **kwargs : field-key pair

        Returns
        -------
        set[T]
            Subset of indexed dataclasses

        Examples
        --------
        >>> from dataclasses import dataclass
        >>> from ktz.dataclasses import Index
        >>> @dataclass(frozen=True)
        ... class A:
        ...     x: int
        ...
        >>> idx = Index(A).add([A(x=1), A(x=2)])
        >>> idx.get(x=2)
        {A(x=2)}

        """
        assert len(kwargs) < 2
        return self.dis(**kwargs)

    def dis(self, **kwargs) -> set[T]:
        """
        ∨ : Obtain unionized subset of queried dataclasses.

        Each indexed field is searched for the associated dataclasses
        and the union of these results is returned.

        Parameters
        ----------
        **kwargs : kwargs
            field-key pairs

        Returns
        -------
        set[T]
            Subset of indexed dataclasses

        Examples
        --------
        >>> from dataclasses import dataclass
        >>> from ktz.dataclasses import Index
        >>> @dataclass(frozen=True)
        ... class A:
        ...     x: int
        ...     y: int
        ...
        >>> idx = Index(A).add([A(x=2, y=1), A(x=2, y=2), A(x=3, y=3)])
        >>> idx.dis(x=2, y=3)
        {A(x=2, y=1), A(x=3, y=3)}

        """
        agg = self._agg(**kwargs)
        return {} if not agg else set.union(*agg)

    def con(self, **kwargs) -> set[T]:
        """
        ∧ : Obtain intersection of queried dataclasses.

        Each indexed field is searched for the associated dataclasses
        and the intersection of these results is returned.

        Parameters
        ----------
        **kwargs : kwargs
            field-key pairs

        Returns
        -------
        set[T]
            Subset of indexed dataclasses

        Examples
        --------
        >>> from dataclasses import dataclass
        >>> from ktz.dataclasses import Index
        >>> @dataclass(frozen=True)
        ... class A:
        ...     x: int
        ...     y: int
        ...
        >>> idx = Index(A).add([A(x=2, y=1), A(x=2, y=2), A(x=3, y=3)])
        >>> idx.con(x=2, y=1)
        {A(x=2, y=1)}

        """
        # as per python complexity wiki, intersection is
        #   O(min(len(s), len(t))
        # so the full set can always be provided even for single element lookups
        # TODO: it can degenerate for multiple sets?
        # https://stackoverflow.com/questions/30845469/time-complexity-of-python-set-intersection-for-n-sets

        agg = self._agg(**kwargs)
        return {} if not agg else set.intersection(*agg)

    def keys(self, field: str) -> set[Any]:
        """
        Obtain indexed keys for the given field.

        This enumerates the possible keys that can be used to retrieve
        indexed dataclasses

        Parameters
        ----------
        field : str
            The field of interest

        Returns
        -------
        set[Any]
            All known keys

        Examples
        --------
        >>> from dataclasses import dataclass
        >>> from ktz.dataclasses import Index
        >>> @dataclass(frozen=True)
        ... class A:
        ...     x: int
        ...
        >>> idx = Index(A).add([A(x=2), A(x=3), A(x=5)])
        >>> idx.keys('x')
        {2, 3, 5}

        """
        if field not in self._idxs:
            self._raise_keyerror(field)

        return set(self._idxs[field])

    def add(self, ts: Iterable[T]) -> "Index":
        """
        Add instances to index.

        Parameters
        ----------
        ts : Iterable[T]
            The dataclass instances to be indexed

        Returns
        -------
        "Index"
            self

        """
        for t, dic in ((t, asdict(t)) for t in ts):
            self._flat.add(t)
            for key in self._idxs:
                self._idxs[key][dic[key]].add(t)

        return self

    def __init__(
        self,
        Klass,
        includes: Iterable[str] = None,
        excludes: Iterable[str] = None,
    ):
        """
        Create a new index.

        Indexes dataclasses of the specified type. Subsets of the
        indexed dataclasses are returned by Index.con and Index.dis.

        Parameters
        ----------
        Klass : T
            Class declaration of the dataclass
        includes : Iterable[str]
            Only include the provided fields
        excludes : Iterable[str]
            Exclude the provided fields

        Raises
        ------
        KeyError
            If includes or excludes contain unknown field names.

        Examples
        --------
        >>> from dataclasses import dataclass
        >>> from ktz.dataclasses import Index
        >>> @dataclass(frozen=True)
        ... class A:
        ...     x: int
        ...
        >>> idx = Index(A).add([A(x=2), A(x=3), A(x=5)])

        """
        includes = set(includes) if includes else set()
        excludes = set(excludes) if excludes else set()

        fieldnames = set(field.name for field in fields(Klass))

        for col in (includes, excludes):
            rem = col - fieldnames
            if rem:
                raise KeyError(f"Unknown fields: {rem}")

        self._idxs = defaultdict(dict)
        for name in fieldnames:

            if includes and name not in includes:
                continue

            if excludes and name in excludes:
                continue

            self._idxs[name] = defaultdict(set)

        self._flat = set()


# ---


class Builder:
    """
    Simple builder to incrementally build a dataclass.

    An instance of this class maintains a kwargs dictionary
    which can be incrementally popularized. Calling the
    instance attempts to build the provided dataclass object.

    Methods
    -------
    __call__:
      Assemble the dataclass
    get:
      Obtain an already added value
    add:
      Add another property

    """

    def __init__(self, Klass, immutable: bool = False):
        """Create a builder.

        Parameters
        ----------
        Klass : dataclass
            Dataclass constructor
        immutable : bool
            whether multiple assignments are allowed

        Examples
        --------
        >>> from dataclasses import dataclass
        >>> from ktz.dataclasses import Builder
        >>> @dataclass
        ... class Product:
        ...     a: int
        ...     b: str
        >>> build = Builder(Klass=Product)
        >>> build.add(a=3)
        >>> build.get('a')
        3
        >>> build.add(b="foo")
        >>> build()
        Product(a=3, b='foo')

        """
        self._Klass = Klass
        self._kwargs = {}
        self._immutable = immutable

    def __call__(self):
        """Assemble the dataclass instance.

        Examples
        --------
        FIXME: Add docs.

        """
        return self._Klass(**self._kwargs)

    def get(self, key: str):
        """Get the value of the provided key.

        Parameters
        ----------
        key : str
        """
        return self._kwargs[key]

    def add(self, **kwargs):
        """Add properties to the dataclass.

        Parameters
        ----------
        **kwargs : Any

        Raises
        ------
        ktz.Error
            If immutable=True, this is raised when a key is provided
            multiple times.

        """
        for key, val in kwargs.items():

            if self._immutable and key in self._kwargs:
                raise ktz.Error(f"cannot overwrite {key}")

            self._kwargs[key] = val



# -*- coding: utf-8 -*-

"""Things to do with the filesystem."""

import ktz

import pathlib
import logging

import git


from typing import Union
from typing import Optional


log = logging.getLogger(__name__)


def path(
    name: Union[str, pathlib.Path],
    create: bool = False,
    exists: bool = False,
    is_dir: Optional[bool] = None,
    is_file: Optional[bool] = None,
    message: str = None,
) -> pathlib.Path:
    """
    Create paths.

    This simply gathers usual operations on Path objects for which
    normally multiple calls to the pathlib are required.

    Parameters
    ----------
    name : Union[str, pathlib.Path]
        The path in question
    create : bool
        Whether to create a directory if it does not exist
    exists : bool
        Check if the path exists, otherwise raise
    is_dir : Optional[bool]
        Check if path is a directory, otherwise raise
    is_file : Optional[bool]
        Check if path is a file, otherwise raise
    message : str
        A message to be logged

    Returns
    -------
    pathlib.Path
        A Path instance

    Raises
    ------
    ktz.Error
        Raised if any of the constraints are violated

    Examples
    --------
    >>> from ktz.filesystem import path
    >>> somedir = path('foo/bar', create=True)
    >>> path(somedir, exists=True)
    PosixPath('foo/bar')
    >>> path(somedir, is_dir=True)
    PosixPath('foo/bar')
    >>> path(somedir, is_file=True)
    Traceback (most recent call last):
      (...)
    Error: foo/bar exists but is not a file

    """
    path = pathlib.Path(name)

    if (exists or is_file or is_dir) and not path.exists():
        raise ktz.Error(f"{path} does not exist")

    if is_file and not path.is_file():
        raise ktz.Error(f"{path} exists but is not a file")

    if is_dir and not path.is_dir():
        raise ktz.Error(f"{path} exists but is not a directory")

    if create:
        path.mkdir(exist_ok=True, parents=True)

    if message:
        path_abbrv = f"{path.parent.name}/{path.name}"
        log.info(message.format(path=path, path_abbrv=path_abbrv))

    return path


def path_rotate(current: Union[str, pathlib.Path], keep: int = None):
    """
    Rotate a file.

    Given a file "foo.tar", rotating it will produce "foo.1.tar".
    If "foo.1.tar" already exists then "foo.1.tar" -> "foo.2.tar".
    And so on. Also works for directories.

    If 'keep' is set to a positive integer, keeps at most
    that much files.

    Parameters
    ----------
    current : Union[str, pathlib.Path]
        Target file or directory
    keep : int
        How many rotated files to keep at most

    Examples
    --------
    >>> from ktz.filesystem import path_rotate
    >>> ! touch test.txt
    >>> path_rotate('test.txt', keep=2)
    >>> ! ls
    test.1.txt
    >>> ! touch test.txt
    >>> ! ls
    test.txt  test.1.txt
    >>> path_rotate('test.txt', keep=2)
    >>> ! ls
    test.1.txt  test.2.txt
    >>> ! touch test.txt
    >>> ! ls
    test.txt test.1.txt test.2.txt
    >>> path_rotate('test.txt', keep=2)
    >>> ! ls
    test.1.txt  test.2.txt

    """
    current = path(current, exists=True)
    if keep:
        assert keep > 1

    def _new(
        p: pathlib.Path,
        n: int = None,
        suffixes: list[str] = None,
    ) -> pathlib.Path:
        name = p.name.split(".")[0]  # .stem returns foo.tar for foo.tar.gz
        return p.parent / "".join([name, "." + str(n)] + suffixes)

    def _rotate(p: pathlib.Path):
        if p.exists():
            old_n, *suffixes = p.suffixes

            n = int(old_n[1:]) + 1
            new = _new(p, n=n, suffixes=suffixes)

            _rotate(new)

            if n <= keep:
                p.rename(new)
            else:
                p.unlink()

    new = _new(current, n=1, suffixes=current.suffixes)
    _rotate(new)
    current.rename(new)


def git_hash() -> str:
    """
    Obtain the current git hash.

    Returns
    -------
    str
        Current git hash

    Examples
    --------
    >>> from ktz.filesystem import git_hash
    >>> git_hash()
    'bedee821a4c1e1217cee783b33ad3bea98dbbb9d'

    """
    repo = git.Repo(search_parent_directories=True)
    # dirty = '-dirty' if repo.is_dirty else ''
    return str(repo.head.object.hexsha)
